/****************************************************************

  Traction Software, Inc. Confidential and Proprietary Information

  Copyright (c) 1996-2015 Traction Software, Inc.
  All rights reserved.

****************************************************************/

// PLEASE DO NOT DELETE THIS LINE -- make copyright depends on it.
package com.traction.view.listener;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import javax.mail.internet.MimeUtility;

import tsi.admin.ServerAdmin;

import com.traction.sdk.AuthenticationException;
import com.traction.sdk.Configuration;
import com.traction.sdk.ConfigurationException;
import com.traction.sdk.Entry;
import com.traction.sdk.EntryOutOfRangeException;
import com.traction.sdk.JournalException;
import com.traction.sdk.TractionId;
import com.traction.sdk.UrlParam;
import com.traction.sdk.User;
import com.traction.sdk.event.Event;
import com.traction.sdk.event.EventTransport;
import com.traction.sdk.iter.ReclassificationIterator;
import com.traction.sdk.mail.AddressException;
import com.traction.sdk.mail.MailConfigurationException;
import com.traction.sdk.mail.Message;
import com.traction.sdk.mail.MessagingException;
import com.traction.sdk.props.GetProperty;
import com.traction.sdk.props.SimpleProperties;
import com.traction.sdk.sdl.CompiledTemplate;
import com.traction.sdk.sdl.Scope;
import com.traction.sdk.sdl.SdlFactory;
import com.traction.sdk.sdl.Tag;
import com.traction.sdk.util.Encoding;
import com.traction.sdk.util.MailUtil;
import com.traction.sdk.util.log.Debug;
import com.traction.sdk.view.TractionView;
import com.traction.view.listener.notifier.transport.EmailTransport;

/**
 * Event listener that sends an email notification to the user's preferred email
 * address, if the appropriate preference is set for that user.
 *
 * There are some useful special behaviors that are activated when event
 * debugging is enabled.  Firstly, a good deal of information about the
 * generated content and the message, including the recipient, is printed to the
 * log file.  Also, the "skipsend" URL parameter may be set to "true" to cause
 * the message to be generated but not sent, which is useful with the listener=
 * parameter when debugging from a web browser.  Note that, as with other
 * listener types, when debugging from a web browser, the event type must be
 * specified manually via the event= URL parameter.
 */
public class EmailNotifier extends TractionView {

    /***************************************************************************
    *                                                                          *
    * Protected static fields.                                                 *
    *                                                                          *
    ***************************************************************************/

    /**
     * The name of the experimental header field that gets set to
     * indicate the type of event that triggered the notification.
     */
    public static final String NOTIFICATION_EVENT_HEADER_NAME = "X-Traction-Notification-Event";

    /***************************************************************************
    *                                                                          *
    * Protected static fields.                                                 *
    *                                                                          *
    ***************************************************************************/
    /**
     * The name of the property in the listener's configuration file that tells
     * the listener which user it should use to run this listener.
     */
    protected static final String RUN_AS_USER_PROPERTY = "user";

    /**
     * The name of the user preference property that tells the listener whether
     * a given user has elected to receive email notifications from this
     * listener for events the user himself triggered.
     */
    protected static final String EMAIL_NOTIFICATION_SELF_PREFERENCE_NAME = "emailnotification_on_own_events";

    /**
     * The name of the user preference property that tells the listener whether
     * a given user has elected to receive email notifications from this
     * listener for events other users triggered.
     */
    protected static final String EMAIL_NOTIFICATION_OTHERS_PREFERENCE_NAME = "emailnotification_on_others_events";

    /**
     * The name of the optional property in the listener's configuration file
     * that may list the users allowed to use this listener.
     */
    protected static final String ALLOW_USERS_LIST_PROPERTY = "users.allow";

    /**
     * The name of the optional property in the listener's configuration file
     * that may list the users are not allowed to use this listener.
     */
    protected static final String DENY_USERS_LIST_PROPERTY = "users.deny";

    /**
     * The name of the optional property in the listener's configuration file
     * that lists the projects for which the listener should send
     * notifications.
     */
    protected static final String LISTEN_PROJECTS_LIST_PROPERTY = "projects.listen";

    /**
     * The name of the optional property in the listener's configuration file
     * that lists the projects for which the listener should skip
     * notifications.
     */
    protected static final String IGNORE_PROJECTS_LIST_PROPERTY = "projects.ignore";

    /**
     * The name of the property in the listener's configuration file that tells
     * the listener which SDL template to use in generating the body of the
     * email message it sends.
     */
    protected static final String MESSAGE_BODY_SDL_TEMPLATE_PROPERTY = "sdltemplate";

    /**
     * The name of the property in the listener's configuration file that tells
     * the listener what email address to use as the "from" field of the email
     * message it sends.
     */
    protected static final String MESSAGE_SENDER_PROPERTY = "sender";

    /**
     * The name of the global variable in the SDL code that stores the event
     * type.
     */
    protected static final String EVENT_TYPE_SDL_VARIABLE_NAME = "event";

    /**
     * The name of the global variable in the SDL code that stores the title
     * to be given to the HTML document the template evaluates to.
     */
    protected static final String HTML_TITLE_SDL_VARIABLE_NAME = "html_title";

    /**
     * For erase and reclassify events, the name of the variable set in the
     * created scope that stores the printable declaration of the effects of
     * the given event.  This takes the form of "Reclassified <Entries |Entry>"
     * followed by a comma separated list of the Traction IDs of the affected
     * entries.  It is intended to be used as the title in the body of the
     * email message sent by this listener.
     */
    protected static final String AFFECTED_ENTRIES_MESSAGE_SDL_VARIABLE_NAME = "affected_entries_list";

    /**
     * For reclassify events, the name of the list variable created in the scope
     * that refers to a Java String[] object containing string representations
     * of the TractionIDs of reclassified entries.  Each value in the list may
     * be used as the "tractionid" attribute of the SDL "entries" tag, so that
     * portions of the affected entries may be generated.
     */
    protected static final String RECLASSIFIED_ENTRIES_SDL_LIST_VARIABLE_NAME = "reclassified_ids";

    /**
     * For reclassify events, the name of the list variable created in the scope
     * that refers to a Java String[] containing string representations of the
     * IDs of the items that were reclassified.  Each element in the list
     * corresponds to the element in the same position in the list of
     * reclassified entries.  Each value in the list may be used as the "skip"
     * attribute of the SDL "items" tag, so that the code may be generate only
     * that item that the reclassification pertains to.
     */
    protected static final String RECLASSIFIED_ITEMS_SDL_LIST_VARIABLE_NAME = "reclassified_items";

    /**
     * For reclassify events, the name of the list variable created in the scope
     * that stores strings, each of which gives a representation of the
     * reclassifications that took place.  Each element in the list corresponds
     * to the element in the same position in the list of reclassified entries.
     * Each element is a string of the following form:
     * "[-:P1:L1[, -:P2:L2, ...]][;][+:P3:L3[, [+:P4:L4, ...]".
     */
    protected static final String LABEL_CHANGES_SDL_LIST_VARIABLE_NAME = "label_changes";

    /**
     * The representation that should be used for a comma literal that is to
     * appear in a comma separated list, so that it will not be confused with
     * the comma that separates items in the list.
     */
    protected static final String COMMA_SEPARATED_LIST_COMMA_LITERAL = "&#x2c;";

    /**
     * The string that should precede any string representation of a label
     * removal.
     */
    protected static final String REMOVED_LABEL_PREFIX = "-";

    /**
     * The string that should precede any string representation of a label
     * addition.
     */
    protected static final String ADDED_LABEL_PREFIX = "+";

    /**
     * The string that should precede the rest of the message subject for
     * certain event notifications to facilitate threading in email readers.
     */
    protected static final String THREADED_MESSAGE_SUBJECT_PREFIX = "re:";

    /***************************************************************************
    *                                                                          *
    * Protected static methods.                                                *
    *                                                                          *
    ***************************************************************************/

    /***************************************************************************
    *                                                                          *
    * Protected instance fields.                                               *
    *                                                                          *
    ***************************************************************************/

    /**
     * This is a shorthand reference for
     * context.getUrl().getUrlParam(UrlParam.EVENT);
     */
    protected String event;

    /**
     * This is a shorthand reference for
     * context.getViewAssembly().getConfiguration().
     */
    protected Configuration config;

    /**
     * The entry that triggered the event.
     */

    protected Entry entry;

    /**
     * The entry that will be used to create the scope.  For an edit event, it
     * will be the entry that is being updated by the edit entry; for an erase
     * event, it will be the entry being erased; for all other events, it will
     * be the same as
     * context.getJournal().getEntry(context.getUrl().getEntry(), jreq).  The
     * scope is intended to be used by an SDL template.
     */
    protected Entry scopeEntry;

    /**
     * The email message that will be sent by this listener.
     */
    protected Message message;

    /**
     * In order to perform the skipNotification check for a reclassify event, we
     * have to retrieve all the reclassifications; in that event, we may as well
     * store them so they don't need to be retrieved again in the setupMessage
     * method.  [shep 09.Mar.2005]
     */
    protected ReclassificationIterator recats;

    /**
     * The recipient name/address actually used to send the messsage.
     */
    protected String recipient;

    /***************************************************************************
    *                                                                          *
    * Public instance methods.                                                 *
    *                                                                          *
    ***************************************************************************/

    /**
     * Generates an email message to notify a given user of a given event.  This
     * method itself a) collects information about the current context, making
     * it available for use in other methods, b) tests whether the notification
     * should proceed, and based on that test may or may not c) create and d)
     * send an email notification to the given user about the given event.
     */
    @Override
    public void generate() {

	// Defined here so as to be available in the catch blocks below.
	String runasuser = "";

 	try {

	    config = context.getViewAssembly().getConfiguration();

	    // If the listener configuration file does not specify a user for
	    // running this listener, throw a ConfigurationException.
	    runasuser = getRunAsUser();

	    event = context.getUrl().getParameter(UrlParam.EVENT);

	    // If an event really triggered this listener, this should never
	    // happen; but if someone is debugging this in a browser, it might
	    // be useful for them to be able to look in the log file and find
	    // out why their attempt to trigger this listener was short-
	    // circuited.
	    if (event == null || event.equals("")) {
		throw new ConfigurationException("Configuration Error: You must have been trying to invoke the email notifer listener from a browser, but you forgot to specify an event= parameter.  Please specify event=[post|edit|reclassify|email|erase|publish|unpublish|lock|unlock] depending which event type you want to simulate.");
	    }

	    Debug.event.println("[" + getListenerName() + "] Received ", event,
				" event; id=",
				context.getUrl().getEntry().toString(),
				"; user=\"", context.getUser().getName(), "\"");

	    String sentId = getTransportId();

	    if (EventTransport.hasAlreadyBeenSent(context, context.getRequest().getEventContext(), sentId)) {
		Debug.event.println("[", getListenerName(), "] skipping email notification (event=",
				    event, "; entry=",
				    context.getUrl().getEntry().toString(),
				    "; recipient=\"", context.getUser().getName(),
				    "\" - \"",
				    getEmailAddress(),
				    "\") SKIPPED because this notification has already been sent.");
		return;
	    }

	    if ( !skipNotification() ) {

		setupMessage();

		boolean skipSendConfig = SimpleProperties.loadBoolean(config, "skipsend", false);
		boolean skipSendUrl = SimpleProperties.loadBoolean(context.getUrl(), "skipsend", false);
		boolean forceSendUrl = SimpleProperties.loadBoolean(context.getUrl(), "forcesend", false);

		boolean skipSend = (skipSendConfig || skipSendUrl) && !forceSendUrl;

		if (skipSend) {
		    Debug.event.println("[", getListenerName(), "] sending email notification (event=",
					event, "; entry=",
					context.getUrl().getEntry().toString(),
					"; recipient=\"", context.getUser().getName(),
					"\" - \"",
					getEmailAddress(),
					"\") SKIPPED");
		} else {
		    Debug.event.println("[", getListenerName(), "] sending email notification (event=",
					event, "; entry=",
					context.getUrl().getEntry().toString(),
					"; recipient=\"", context.getUser().getName(),
					"\" - \"",
					getEmailAddress(),
					"\")...");
		    message.send();
		    EventTransport.recordSent(context, context.getRequest().getEventContext(), sentId);
		}

		Debug.log.println("[" + getListenerName() + "] " + event + " event notification for " +
				  context.getUrl().getEntry().toString() + " sent to " +
				  recipient + " (" + context.getUser().getName() + ")"+(skipSend ? " SKIPPED" : ""));

	    }

	} catch (ConfigurationException ce) {
	    Debug.event.println("[" + getListenerName() + "] ", ce.getMessage());
 	} catch (AddressException ae) {
 	    Debug.event.println("[" + getListenerName() + "] ", ae.getMessage());
 	} catch (AuthenticationException xcp) {
	    Debug.event.println(xcp.getMessage());
	    Debug.event.println("[" + getListenerName() + "] This listener is running as user \"", runasuser,
				"\", but that user can't read the entry.");
 	    Debug.event.println("[" + getListenerName() + "] Please change the value of the user property ",
				"in the listener's .properties file.");
 	} catch (JournalException xcp) {
 	    Debug.event.println(xcp.getMessage());
 	} catch (IOException ioe) {
 	    Debug.event.println(ioe.getMessage());
 	} catch (Exception e) {
	    String min_context = null;
	    // Make an effort to provide context for this unexpected problem.
	    // Do it incrementally; in case one of these operations causes an
	    // exception, we want to accumulate as much information as possible.
	    try {
		min_context  = "(";
		min_context += "event=" + event;
		min_context += "; recipient=\"" + context.getUser().getName();
		min_context += "\"; entry=" + context.getUrl().getEntry().toString() + ")";
	    } catch (Exception other) {
		if (min_context.equals("(")) {
		    min_context = "(no context available)";
		} else {
		    min_context += ")";
		}
	    }
 	    Debug.log.println("[" + getListenerName() + "] Unexpected error in EmailNotifier " +
			      min_context + ":  " + e.getMessage());
	    e.printStackTrace(Debug.log);
 	}

    }

    protected final String getTransportId() {
	return "email-" + getEmailAddress();
    }

    protected String getEmailAddress() {
	return context.getUser().getUserPreferences().getPreferredEmail();
    }

    public String getRunAsUser() {
	String runasuser = SimpleProperties.loadString(config, RUN_AS_USER_PROPERTY, null);
	if (runasuser == null) {
	    throw new ConfigurationException("Configuration Error: This listener requires a user property, " +
					     "but none is specified, or the property has no value.");
	}
	return runasuser;
    }

    @Override
    public int getVolumeIndex() {
	return VOL_IGNORE;
    }

    /***************************************************************************
    *                                                                          *
    * Protected instance methods.                                              *
    *                                                                          *
    ***************************************************************************/

    /**
     * Tests whether notification is allowed to proceed, according to various
     * rules.  If the listener configuration file is misconfigured; or if its
     * configuration does not allow the current user to receive notification; or
     * if the  listener's configuration specifies that events in the given
     * project are to be ignored; or if the given user has not asked to receive
     * notification, notification should be skipped.
     * @return false if any of the conditions above holds; true otherwise.
     * @throws AuthenticationException if the entry from the context (and/or, in
     * in the case of an edit or reclassify event, the entry to which that entry
     * refers) is not readable by the user that the listener is running as.
     * @throws EntryOutOfRangeException if the entry from the context
     * (and/or, in the case of an edit or reclassify event, the entry to which
     * that entry refers) is not readable by the user that the listener is
     * running as.
     */
    protected boolean skipNotification()
	throws AuthenticationException, JournalException {

        if (Notifier.skipEmailRecordNoContent(context)) {
            return true;
        }
        if (Notifier.skipReclassifyRecordNoVisibleLabels(context)) {
            return true;
        }

	// If _either_ of the following holds for the user who is supposed to
	// receive notification, then the notification should be skipped:
	// 1) the user is the author of this entry (i.e., the user triggered
	// this listener) but the user has opted out of receiving notifications
	// about events that he himself triggered; 2) the user isn't the author
	// of this entry (i.e., some other user triggered this listener) but
	// the user has opted out of receiving notifications about events that
	// other users triggered.
	User user = context.getUser();
	String username = user.getName();
	GetProperty prefs = user.getUserPreferences().getProperties(context);
	boolean notifyauthorself = false;
	boolean notifyauthorothers = false;
	try {
	    notifyauthorself = prefs.getProperty(EMAIL_NOTIFICATION_SELF_PREFERENCE_NAME).equals("true");
	    notifyauthorothers = prefs.getProperty(EMAIL_NOTIFICATION_OTHERS_PREFERENCE_NAME).equals("true");
	} catch (NullPointerException e) {
	    // do nothing
	}

	// So that we don't have to make a request to the journal to retrieve
	// the entry (in order to get the author) right now, we should do this
	// check up front:  if the user has opted out of notification whether or
	// not the user is the author, we can stop right here.  This may speed
	// things up when skipping notification.  [shep 02.Sep.2005]
	if (!notifyauthorself && !notifyauthorothers) {
	    Debug.event.println("[" + getListenerName() + "] Skipping email notification for user \"", username,
				"\" because that user has opted out of receiving notifications",
				" for events (triggered by self or others).");
	    return true;
	}

	TractionId id = context.getUrl().getEntry();

	// If we try to get the entry and can't, it means that the listener is
	// mis-configured because the user property is missing altogether, has
	// no value, or specifies an unrecognized user or a user who does not
	// have permission to read the entry; or that the id is out of range
	// (the entry does not exist).
	entry = context.getJournal().getEntry(id, jreq);

	boolean isauthor = entry.isMine();
	if (isauthor) {
	    if (!notifyauthorself) {
		Debug.event.println("[" + getListenerName() + "] Skipping email notification for user \"", username,
				    "\" because that user has opted out of receiving notifications",
				    " for events he/she triggered.");
		return true; // user doesn't want notification of an entry he/she authored
	    }
	}
	else {
	    if (!notifyauthorothers) {
	    Debug.event.println("[" + getListenerName() + "] Skipping email notification for user \"", username,
				"\" because that user has opted out of receiving notifications",
				" for events other users triggered.");
	    return true; // user doesn't want notification of an entry another user authored
	    }
	}

	// If the entry is an edit event, use the entry that is being
	// updated to create the body; else use the given entry itself.  If this
	// fails with an exception, it means that the entry id of the updated
	// entry is out of bounds or more likely that the entry being updated
	// cannot be read by the user that the listener is running as.
	if (event.equals(Event.EDIT)) {
	    scopeEntry = context.getJournal().getEntry(entry.getEntryUpdated(), jreq);
	} else {
	    scopeEntry = entry;
	}
	Debug.event.println("[" + getListenerName() + "] Using ",
			    scopeEntry.getTractionId().toString(),
			    " as the entry for the message body of notification for ", event,
			    " event, entry ", entry.getTractionId().toString(), ".");

	// Make sure the user is on the allow list, if there is one, and is not
	// on the deny list, if there is one.
	String allowusers = SimpleProperties.loadString(config, ALLOW_USERS_LIST_PROPERTY, null);
	if (allowusers != null) {
	    if (allowusers.indexOf("|" + username + "|") < 0) {
		Debug.event.println("[" + getListenerName() + "] Skipping ", event,
				    " notification for entry ", id.toString(),
				    " for user \"", username,
				    "\" because that user is not on the allow list.");
		return true; // this user is not on the allow list
	    }
	}
	String denyusers = SimpleProperties.loadString(config, DENY_USERS_LIST_PROPERTY, null);
	if (denyusers != null) {
	    if (denyusers.indexOf("|" + username + "|") > -1 ) {
		Debug.event.println("[" + getListenerName() + "] Skipping ", event,
				    " notification for entry ", id.toString(),
				    " for user \"", username,
				    "\" because that user is on the deny list.");
		return true; // this user on the deny list
	    }
	}

	// Make sure the project is on the listen list, if there is one, and
	// not on the ignore list, if there is one.
	String project = id.getProject().getName();
	String updateprojects = SimpleProperties.loadString(config, LISTEN_PROJECTS_LIST_PROPERTY, null);
	if (updateprojects != null) {
	    if (updateprojects.indexOf("|" + project + "|") < 0) {
		Debug.event.println("[" + getListenerName() + "] Skipping ", event,
				    " notification for entry ", id.toString(),
				    " because its project is not on the listen list.");
		return true; // this project is not on the allow list
	    }
	}

	String skipprojects = SimpleProperties.loadString(config, IGNORE_PROJECTS_LIST_PROPERTY, null);
	if (skipprojects != null) {
	    if (skipprojects.indexOf("|" + project + "|") > -1 ) {
		Debug.event.println("[" + getListenerName() + "] Skipping ", event,
				    " notification for entry ", id.toString(),
				    " because its project is on the ignore list.");
		return true; // this project is on the skip list
	    }
	}

	// If all tests succeeded, return false:  do not skip the notification.
	Debug.event.println("[" + getListenerName() + "] Notification will proceed (event=", event,
			    "; entry=", context.getUrl().getEntry().toString(),
			    "; recipient=\"", username, "\").");
	return false;

    }

    /**
     * This method constructs the body and headers of the email message to be
     * sent by this listener, using the SDL template specified in the
     * listener's configuration file.
     * @throws ConfigurationException if the listener configuration files does
     * not specify an "sdltemplate" property, or if the SDL template specified
     * does not exist or is otherwise invalid (as determined by
     * context.getSdlFactory().createTemplate()); if the listener configuration
     * file does not specify a "sender" property, or if the address specified is
     * invalid; or if the user's preferred email address is invalid; if the user
     * being used to run the listener does not have permission to send mai; or
     * if there is a problem with this Traction Server's mail configuration.
     * @throws IOException if there is an I/O problem while evaluating the SDL
     * template; or if the entry's title cannot be retrieved.
     * @throws UnsupportedEncodingException if there is a problem converting the
     * body to a String with the same encoding as Traction's internal encoding
     * (UTF-8).
     */
    protected void setupMessage()
	throws ConfigurationException, IOException, UnsupportedEncodingException {

	// Check what message body type is desired, and derive from that the
	// name of the SDL template to be used to generate the message body.
	String sdltemplate = SimpleProperties.loadString(config, MESSAGE_BODY_SDL_TEMPLATE_PROPERTY, null);
	if (sdltemplate == null) {
	    throw new ConfigurationException("Configuration Error: Configuration '" +
					     config.getName() +
					     "' is missing the property " +
					     MESSAGE_BODY_SDL_TEMPLATE_PROPERTY + ".");
	}
	Debug.event.println("[" + getListenerName() + "] Using SDL template ", sdltemplate);

	// Try to create the template, and make sure creation succeeded.
	SdlFactory sdl = context.getSdlFactory();
	CompiledTemplate template = sdl.createTemplate(sdltemplate);
	if (template == null) {
	    throw new ConfigurationException("Configuration Error: Configuration '" +
					     config.getName() + "' property " +
					     MESSAGE_BODY_SDL_TEMPLATE_PROPERTY +
					     " refers to non-existant template '" +
					     sdltemplate + "'.");
	}

	// Prepare to evaluate the SDL template for the message body
	Tag root = template.getRootBlock(sdl);
	Scope scope = sdl.createScope(template, context, root);
	scope.setCurrentEntry(scopeEntry);
	scope.setVariable("originalid", entry.getTractionId().toString());
	scope.setGlobalVariable(EVENT_TYPE_SDL_VARIABLE_NAME, event);

        if (event.equals(Event.EDIT)) {
            modifyScopeForEdit(scope);
	}
        if (event.equals(Event.RECLASSIFY)) {
            modifyScopeForReclassify(scope);
        }

	// Try to create the message.
	try {
	    message = context.getSdkFactory().createSmtpMessage(context);
	}
	catch (AuthenticationException ae) {
	    throw new ConfigurationException("Configuration Error: The user \"" +
					     SimpleProperties.loadString(config, RUN_AS_USER_PROPERTY, null) +
					     "\" being used to run this listener does not have permission to send mail.");
	}
	catch (MailConfigurationException mce) {
	    throw new ConfigurationException("Configuration Error: " + mce.getMessage());
	}

	// Optionally use the project specific friendly name.

	// Retrieve the sender address from the listener configuration.  If the
	// address is not specified, retrieve the digest sender address from
	// Server Setup | Email.  If that address is also not specified, throw a
	// ConfigurationException.  Otherwise, use it as the sender's address
	// for this message (From: field).
	String sender = SimpleProperties.loadString(config,
						    MESSAGE_SENDER_PROPERTY,
						    ServerAdmin.getServerDataInterpreter().getDigestSender());
	if (sender == null || sender.trim().equals("")) {
	    throw new ConfigurationException("Configuration Error: No sender specified in " +
					     "email notifier listener's configuration file, " +
					     "and no digest sender address was present to use "+
					     "instead.");
	}

	if (sender.indexOf("@") == -1) {
	    throw new ConfigurationException("Configuration Error: Sender email address (\"" +
					     sender + "\") does not appear to be a valid email address.");

	}

	try {
	    message.addFrom(sender);
	} catch (AddressException ae) {
	    throw new ConfigurationException("Configuration Error: Sender email address (\"" +
					     ae.getAddress() + "\") is invalid.\nOriginal Message:  " +
					     ae.getMessage());
	}

	// Retrieve the user's preferred email address from the user's
	// preferences.  If the user does not have a preferrred email address,
	// or if the address is invalid, throw a ConfigurationException.  If the
	// address is valid, use it as the recipient address for this message
	// (To: field).
	//
	// Include the user's full name in the recipient address, as the
	// friendly name.  And encode it in a mail-safe fashion.
	recipient = getEmailAddress();
	if (recipient == null || recipient.trim().equals("")) {
	    throw new ConfigurationException("Configuration Error: Cannot notify "
					     + context.getUser().getName() + " of " + event +
					     " event (" + context.getUrl().getEntry().toString() +
					     ") because this user has no preferred email address.");
	}
	else {
	    String fullname = context.getUser().getUserPreferences().getFullName();
	    if (fullname != null && !fullname.trim().equals("")) {
		try {
		    recipient = MailUtil.makeFriendlyNameAddress(recipient, MimeUtility.encodeText(fullname, "utf-8", null));
		} catch (UnsupportedEncodingException uee) {
		    uee.printStackTrace(Debug.log);
		}
	    }
	}

	try {
	    message.addRecipient(recipient, Message.ADDRTYPE_TO);
	}
	catch (AddressException ae) {
	    throw new ConfigurationException("Configuration Error: Preferred email address for user " +
					     context.getUser().getName() + " (\"" +
					     ae.getAddress() + "\") is invalid.\nOriginal Message:  " +
					     ae.getMessage());
	}

	handleSubjectAndBody(scope, root);

    }

    protected void modifyScopeForEdit(Scope scope) {
        EmailTransport.modifyScopeForEdit(scope, entry);
    }

    protected void modifyScopeForReclassify(Scope scope) {
        EmailTransport.modifyScopeForReclassify(context, scope, entry);
    }

    protected void handleSubjectAndBody(Scope scope, Tag bodyBlock) throws IOException, UnsupportedEncodingException {

	String subject = EmailTransport.getSubjectForEntry(context, event, scopeEntry);

	// Finally we can set the title of the HTML document produced by the SDL
	// template.  Just use the subject of the message.
	scope.setGlobalVariable(HTML_TITLE_SDL_VARIABLE_NAME, subject);

	// The output from the template evaluation will go to baos, which has to
	// be wrapped in a TractionOutputStream to be passed to the eval method.
	ByteArrayOutputStream baos = new ByteArrayOutputStream();
	bodyBlock.eval(scope, context.getSdkFactory().createOutputStream(baos));

	subject = EmailTransport.getModifiedSubjectForTargetId(subject, event, scopeEntry, context);

	EmailTransport.setSubject(message, subject);

	try {
	    message.setHeader(NOTIFICATION_EVENT_HEADER_NAME, event + " " + entry.getTractionId());
	}
	catch (MessagingException e) {
	    Debug.logWarning(Debug.event, e, "Failed to set " + NOTIFICATION_EVENT_HEADER_NAME + " header on this notification message");
	}

	// Set the body of the message.  Don't forget to convert the stream with
	// the correct encoding.
	String bodyStr = baos.toString(Encoding.getInternalEncoding());
	message.setBodyHtml(bodyStr);

	// If debug is enabled, send output to the context output stream also
	// (i.e., back to the browser that may be invoking the view).
	if (Debug.event.enabled()) {
	    context.getOutputStream().print(bodyStr);
	}

    }

    protected String getListenerName() {
	return context.getViewAssembly().getConfiguration().getName();
    }

}
